# David Leadbeater, dgl@dgl.cx, GPLv2
use strict;
use Irssi;
use IO::Socket;
use POSIX;
use Fcntl;

# Stores the input tag and filehandle of the listener
my($listener, $listenfh);
# socket is a filehandle to input tag map, buffer is the data read from each
# input handle that isn't a complete line
my(%socket, %buffer);
# Commands usable over the socket not Irssi commands, command registered by
# scripts (See command_add).
my %command;

# Open the Unix domain listener socket
sub listen_new {
	my $path = Irssi::settings_get_str("socket_path");
	unlink $path if -S $path and -o $path;

	$listenfh = IO::Socket::UNIX->new(
		Listen => SOMAXCONN,
		Local  => $path,
		Reuse  => 1
	) or die "Error opening listen socket: $!";
	
	fcntl($listenfh, O_NONBLOCK, 1);

	$listener = Irssi::input_add(fileno $listenfh, INPUT_READ, \&listen_accept,
		$listenfh);
}

# Removes and closes the listener socket
sub listen_close {
	Irssi::input_remove($listener);
	close $listenfh;
	unlink Irssi::settings_get_str("socket_path");
}

# Accepts a new connection on the listener socket and adds it
# Called by the Irssi input system
sub listen_accept {
	my($fh) = @_;
	my $new = $fh->accept;
	return unless ref $new;
	fcntl($new, O_NONBLOCK, 1);
	$socket{$new} = Irssi::input_add(fileno $new, INPUT_READ, \&input_in, $new);
}

# Gets the input and (hopefully) does something useful with it
# Called by the Irssi input system
sub input_in {
   my($fh) = @_;
	my $ret = sysread($fh, $buffer{$fh}, POSIX::BUFSIZ);

	# undef = error (except when it would block), 0 = eof
	if((not defined $ret and $! != POSIX::EAGAIN) or $ret == 0) {
	   input_close($fh);
		return;
	}

	return unless $buffer{$fh} =~ /\n/;

	while($buffer{$fh} =~ s/^(.*?)\r?\n//g) {
		my($command, $data) = split / /, $1, 2;
		command_run(lc $command, $fh, $data);
	}
}

sub input_out {
	my($fh, $data) = @_;
	syswrite($fh, $data, length $data);
}

# Removes and closes an input handle
sub input_close {
	my($fh) = @_;
	Irssi::input_remove($socket{$fh});
	close $fh;
	
	delete $socket{$fh};
	delete $buffer{$fh};
}

# Adds a new command, accepts a reference to a subroutine that will be passed
# the filehandle, data after the command and the command name itself as
# arguments.
sub command_add {
	my($command, $ref) = @_;
	$command{lc $command} = $ref;
}

# Called by input_in to actually run the command.
sub command_run {
	my($command, $fh, $data) = @_;
	if(not exists $command{$command}) {
		print "Attempt to run unknown command '$command' over unix socket";
		return;
	}
	$command{$command}->($fh, $data, $command);
}

# Example command..
command_add "close" => sub {
	my($fh, $data) = @_;
	input_close($fh);
};

Irssi::settings_add_str("misc", "socket_path", Irssi::get_irssi_dir . "/sock");
listen_new();
1; # vim:set noexpandtab:
